//Title: MK MediaPlayer Main Window Class
//Platform: BeOS 5
//Version: 2002.03.19
//Description:
//	A class that creates the application's "main" window.
//	This class inherits from BWindow, and runs a message loop for the window.
	
//Copyright (C) 2001, 2002 Kevin H. Patterson
//Available for use under the terms of the BSD license agreement:
//See the associated file "License.txt" for details.

#ifndef _clsMainWindow_h
#define _clsMainWindow_h
	
	#include <Application.h>
	#include <View.h>
	#include <Window.h>
	#include <Messenger.h>
	#include <FilePanel.h>
	#include <MenuBar.h>
	#include <Slider.h>
	#include <MessageRunner.h>
	#include <String.h>

	#include "globals.h"
	
	#include "MediaPlayerNode.h"
	
	struct MediaConnection {
		media_node_id Producer;
		media_node_id Consumer;
		media_output Output;
		media_input Input;
	};
	
	class clsMainWindow
	:
		public BWindow
	{
	public:
		clsMainWindow( const char *uWindowTitle );
		~clsMainWindow();
		virtual void MessageReceived( BMessage *Message );
		virtual bool QuitRequested();
	
	private:
		BMessenger * iMessenger;
		BFilePanel * iOpenFilePanel;
		BMenuBar * iMenuBar;
		BSlider * iSeekSlider;
		BMessageRunner * iUpdateTimer;
		BString FileName;
	
		MediaPlayerNode * iMediaPlayerNode;
		BTimeSource * iTimeSourceNode;
		media_node MediaProducer;
		media_node AudioConsumer;
		media_node VideoConsumer;
		media_node TimeSource;
		int32 ConnectionCount;
		MediaConnection Connection[64];
		bool UsesAudio;
		bool UsesVideo;
		
	private:
		status_t ConstructConsumerNodes();
		status_t ConstructProducerNode( const char * uFileName );
		status_t Connect();
		status_t Start();
		status_t Stop();
		status_t Disconnect();
		status_t DestructAllNodes();
		status_t DestructProducerNode();
	};
	
#endif
