/*
** Copyright 2001, Michael Pfeiffer	
*/

#include "../include/DriverInterface.h"
#include "../include/AccelerantInterface.h"
#include <unistd.h>
#include <stdio.h> /* for sprintf */

/* private */

extern int driver_fd; /* defined in accelerant.c */

int get_int(char* label, int dfault) {
	setting_int s;
	char a[80];
	sprintf(a, "%d", (int)dfault);
	s.label  = label;
	s.dfault = a;
	if (B_OK == ioctl(driver_fd, GDW_GET_INT, &s, sizeof(s))) {
		return s.result;
	} else {
		return dfault;
	}
}

bool get_bool(char* label, bool dfault) {
	setting_bool s;
	s.label  = label;
	s.dfault = dfault;	
	if (B_OK == ioctl(driver_fd, GDW_GET_BOOL, &s, sizeof(s))) {
		return s.result;
	} else {
		return dfault;
	}
}

/* public */

/* hertz */
bool is_hertz_enabled() {
	return get_bool("changerefreshrate", false);
}

static char *fs_modes[] = {
	"refreshrate640x480",
	"refreshrate800x600",
	"refreshrate1024x768",
	"refreshrate1152x900",
	"refreshrate1280x1024",
	"refreshrate1600x1280"
};

int get_fullscreen_refreshrate(enum full_screen_mode mode) {
	if (mode >= 0 && mode < NO_OF_FULLSCREEN_MODES) {
		return get_int(fs_modes[mode], 600);
	} else {
		return 600;
	}
}

/* BeTVOut */
bool is_tvout_enabled() {
	return get_bool("enabletvout", false);
}

static char *tvout_mode[] = {
	"XXXunusedXXX",
	"dualmode",
	"tvoutonly"
};

static int tvout_rr[] = {
	0,
	62,
	63
};

int  get_tvout_refreshrate(enum tv_out_mode mode) {
	if (mode >= 1 && mode < NO_OF_TV_OUT_MODES) {
		return get_int(tvout_mode[mode], tvout_rr[mode]);
	}
	return 0;
}

