/*
** Copyright 2001, Michael Pfeiffer	
*/

#if !defined(ACCL_INTERFACE_H)
#define ACCL_INTERFACE_H

#include <Accelerant.h>

#if defined(__cplusplus)
extern "C" {
#endif

#ifdef DEBUG
#include <unistd.h>
extern  int log_fd;
#define LOG(s)    { if (log_fd != -1) write(log_fd, s, strlen(s)); }
#define START_LOG { log_fd = open(log_path, O_RDWR | O_CREAT | O_TEXT | O_TRUNC, 0333);}
#define STOP_LOG  { close(log_fd); log_fd = -1; }
#else
#define LOG(s)    {}
#define START_LOG {}
#define STOP_LOG  {}
#endif

/* Entry to accelerant */
extern void *  get_accelerant_hook(uint32 feature, void *data);

/* Entry to wrapped accelerant of original graphics driver */
extern void * (*a_get_accelerant_hook)(uint32 feature, void *data);

/* Function pointers to wrapped accelerant */
extern status_t (*A_INIT_ACCELERANT)(int fd);
extern void     (*A_UNINIT_ACCELERANT)(void);
extern status_t (*A_CLONE_ACCELERANT)(void *data);
extern status_t (*A_SET_DISPLAY_MODE)(display_mode *mode_to_set);
extern status_t (*A_GET_DISPLAY_MODE)(display_mode *mode);

/* Our implementation */
status_t INIT_ACCELERANT(int fd);
status_t CLONE_ACCELERANT(void *data);
void     UNINIT_ACCELERANT(void);
status_t SET_DISPLAY_MODE(display_mode *mode_to_set);
status_t GET_DISPLAY_MODE(display_mode *mode);

/* Fullscreen mode hack */
extern status_t hack_display_mode(display_mode *mode_to_set);

/* path where tv-out add-ons are searched (relative to system/user add-on path) */
#define TV_OUT_ADD_ON_PATH "gdw/tvout/"

/* tv-out prototypes */
extern void load_tv_out_add_on(int driver_fd);
extern void unload_tv_out_add_on();
extern status_t tv_out_set_display_mode(display_mode* mode_to_set, 
                                        status_t (*set_disp_mode)(display_mode*));
extern status_t tv_out_get_display_mode(display_mode* mode);

/* settings */

/* get_int and get_bool should not be used by add-ons*/
extern int get_int(char* label, int dfault);
extern bool get_bool(char* label, bool dfault);

/* add-ons should use these functions: */
extern bool is_hertz_enabled();

enum full_screen_mode {
	MODE640x480 = 0,
	MODE800x600,
	MODE1024x768,
	MODE1152x768,
	MODE1280x1024,
	MODE1600x1200,
	NO_OF_FULLSCREEN_MODES
};
extern int  get_fullscreen_refreshrate(enum full_screen_mode mode); /* in 1/10 Hz */

extern bool is_tvout_enabled();

enum tv_out_mode {
	VGA_MODE,
	DUAL_MODE,
	TV_OUT_MODE,
	NO_OF_TV_OUT_MODES
};

extern int  get_tvout_refreshrate(enum tv_out_mode mode); /*  in 1/10 Hz */

#if defined(__cplusplus)
}
#endif

#endif
