/*
** Copyright 2001, Michael Pfeiffer	
*/

#include "common.h"

#ifdef DEBUG
#include <unistd.h>
#include <stdio.h>
int  log_fd = -1;
char *log_path = "/boot/var/log/tvout-addon";
#endif

int setVESA640 (set_display_mode_t *set_display_mode, display_mode new_mode);
int setVESA800 (set_display_mode_t *set_display_mode, display_mode new_mode);

/* global data */
int32             api_version = TV_OUT_ADDON_API_VERSION;
int               riva_iic_fd = -1;
display_mode      vga_mode;
enum tv_standard  tv_standard;   
enum riva_version riva_version; 
byte              use_ffilter;
int               driver_fd;

int get_int(char* label, int dfault) {
	setting_int s;
	char a[80];
	sprintf(a, "%d", (int)dfault);
	s.label  = label;
	s.dfault = a;
	if (B_OK == ioctl(driver_fd, GDW_GET_INT, &s, sizeof(s))) {
		return s.result;
	} else {
		return dfault;
	}
}

bool get_bool(char* label, bool dfault) {
	setting_bool s;
	s.label  = label;
	s.dfault = dfault;	
	if (B_OK == ioctl(driver_fd, GDW_GET_BOOL, &s, sizeof(s))) {
		return s.result;
	} else {
		return dfault;
	}
}
enum tv_standard _EXPORT get_tv_standard() {
	return (enum tv_standard) get_int("tvstandard", PAL);
}

enum riva_version _EXPORT get_riva_version() {
	return (enum riva_version) get_int("rivamodel", TNT1);
}

enum flicker_filter _EXPORT get_flicker_filter() {
	return (enum flicker_filter) get_int("useffilter", enable_ffilter);
}

void load_settings() {	
	// load settings (like PAL or NTSC, filters, ...)
	use_ffilter  = get_flicker_filter();
	riva_version = get_riva_version();
	tv_standard  = get_tv_standard();
	load_tv_encoder_settings();
}

bool open_riva(int* fd, int bus_found[2]) {
	char iicversion = 0; 
	int bus = 0;
	int bus_wired[2];
	*fd = open("/dev/video/riva_iic/1", O_RDWR);
	//check if driver could be reached...
	if (*fd < 0) {
		LOG("Could not reach 'riva_iic' driver, exiting...\n");
		return false;
	}

	//check version of driver...
	if (ioctl(*fd, rivaVersion, &iicversion)) {
		LOG("Incompatible driver version (V0.12 or higher needed), exiting...\n");
		goto error;
	}
	if (iicversion < 12) {
		LOG("Incompatible driver version (V0.12 or higher needed), exiting...\n");
		goto error;
	}
	// printf("riva_iic driver V%.2f\n",iicversion[0]/100.0);

	// unlock RIVA extended VGA registers...
	ioctl(*fd, rivaUnlock);

	//check for one or two available (wired) IIC busses.
	//On RIVA TNT and later, two could be there...
	// RIVA IICbus 0:
	bus = 0;
	bus_wired[0] = !ioctl(*fd, rivaChkBus, &bus);
	// RIVA IICbus 1:
	bus = 1;
	bus_wired[1] = !ioctl(*fd, rivaChkBus, &bus);
	
	if (!(bus_wired[0] || bus_wired[1])) {
		LOG("No valid RIVA config found, exiting...\n");
		goto error;
	}
	
	//find at least one sane IIC bus... (test this by sending an unused adress...)
	if (FindSaneBus (*fd, bus_wired[0], bus_wired[1], &bus_found[0], &bus_found[1])) {
		return true;
	}
error:
	close(*fd);
	return false;
}


int setVESA640 (set_display_mode_t *set_display_mode, display_mode new_mode) {
	LOG("setVESA640\n");
	//set VESA640: (hex values are exact copies of direct register programmings)
	new_mode.virtual_width = 640;
	new_mode.virtual_height = 480;
	new_mode.timing.h_display = 640;
	new_mode.timing.v_display = 480;
	
	new_mode.timing.h_sync_start = (0x53 * 8);//byte oriented!
	new_mode.timing.h_sync_end = ((0x1f + (0x53 && 0xe0)) * 8);//byte oriented!
	new_mode.timing.h_total = (0x5F * 8);//byte oriented! (val. *must* be set!)
	new_mode.timing.v_sync_start = 0x1ea;
	new_mode.timing.v_sync_end = ((0x1ea && 0xff0) + 0x00c);
	new_mode.timing.v_total = 0x20b;
	new_mode.timing.pixel_clock = ((((0x5F * 8) * 0x20b)* 60)/1000);// in kHz!

	return set_display_mode(&new_mode);
}

int setVESA800 (set_display_mode_t *set_display_mode, display_mode new_mode) {
	LOG("setVESA800\n");
	//set VESA800: (hex values are exact copies of direct register programmings)
	new_mode.virtual_width = 800;
	new_mode.virtual_height = 600;
	new_mode.timing.h_display = 800;
	new_mode.timing.v_display = 600;
	
	new_mode.timing.h_sync_start = (0x6a * 8);//byte oriented!
	new_mode.timing.h_sync_end = ((0x1a + (0x6a && 0xe0)) * 8);//byte oriented!
	new_mode.timing.h_total = (0x7f * 8);//byte oriented! (this value *must* be set!)
	new_mode.timing.v_sync_start = 0x259;
	new_mode.timing.v_sync_end = ((0x259 && 0xff0) + 0x00d);
	new_mode.timing.v_total = 0x272;
	new_mode.timing.pixel_clock = (((((0x7F) * 8) * 0x272)* 60)/1000);// in kHz!

	return set_display_mode(&new_mode);
}

extern "C" status_t _EXPORT load_tv_out(int fd) {
	int bus[2];
	START_LOG
	LOG("load_tv_out\n");
	driver_fd = fd;
	load_settings();
	if (open_riva(&riva_iic_fd, bus)) {
		LOG("riva tv encoder\n");
		if (detect_tv_encoder(riva_iic_fd, bus)) {
			return B_OK;
		}
		close(riva_iic_fd); riva_iic_fd = -1;
	}
	LOG("load_tv_out failed\n");
	return B_ERROR;
}

extern "C" void _EXPORT unload_tv_out(void) {
	LOG("unload_tv_out\n");
	// close riva_iic file
	if (riva_iic_fd != -1) {
		close(riva_iic_fd); riva_iic_fd = -1;
	}
	STOP_LOG
}

extern "C" status_t _EXPORT set_tv_out_mode(display_mode *dm, enum tv_out_mode mode, set_display_mode_t *set_display_mode) {
	LOG("set_tv_out_mode entered\n");

	if (mode == VGA_MODE) {
		LOG("VGA_MODE\n");
		// turn off tv out/dual mode and turn on vga mode
		switch_to_vga(set_display_mode); 
		return B_OK; // note: this return value is ignored
	} else {
		// check if resolution is supported; if not return B_ERROR
		// informations can be obtained from:
		//   dm->space
		//   dm->timing.h_display
		//   dm->timing.v_display
		load_settings();
		if (mode == DUAL_MODE) { LOG("DUAL_MODE\n"); } else { LOG("TV_OUT_MODE\n"); }
		if (is_mode_supported(mode, dm->space, dm->timing.h_display, dm->timing.v_display)) {
			// set tv out mode (either tv-out (mode == TV_OUT_MODE) or 
			// dual mode (mode == DUAL_MODE))
			if (set_mode(dm, mode, set_display_mode))
				return B_OK;			
		}
	}
	return B_ERROR;
}

