/*
** Copyright 2001, Michael Pfeiffer	
*/

#include "common.h"

/* 
   TODO: 
   * settings.c should not be linked with this add-on instead it 
     should be linked with wrapper.accelerant.
*/

typedef struct {
	int16 width;
	int16 height;
	enum tv_standard tv_standard;
	int  input;
	int  ioctl_fct;
	int (*initBT)(int fd);
	int (*setMode)(set_display_mode_t *set_display_mode, display_mode cur_mode);
} tv_mode_t;

/* brooktree */
static int               bt_version;
static int               bt_type;
  
static resolution supported_resolution[] = {
	{640, 480},
	{800, 600},
	{-1, -1}
};

static tv_mode_t tv_mode[] = {
	{640, 480, NTSC, 2, rivaNtsc640, InitBT_NTSC640, setVESA640},
	{800, 600, NTSC, 3, rivaNtsc800, InitBT_NTSC800, setVESA800},
	{640, 480, PAL , 5, rivaPal640,  InitBT_PAL640,  setVESA640},
	{800, 600, PAL , 6, rivaPal800,  InitBT_PAL800,  setVESA800},
// XXX: add DVD modes
	{-1,   -1, PAL , 0, 0,           NULL,           NULL}
};

bool detect_tv_encoder(int fd, int bus_found[2]) {
	LOG("detect_brooktree entered\n");
	int btfound = FindBT(fd);
	bt_version = 0xff; bt_type = 0xff;
	if (btfound && ReadType(fd, &bt_version, &bt_type) && bt_type >= 0 && bt_type <= 1) {
		return true;
	}
	LOG("detect_brooktree failed\n");
	return false;
}


void switch_to_vga(set_display_mode_t *set_display_mode) {
	LOG("switch_to_vga\n");
	LOG("switch_to_vga done\n");
}

bool is_mode_supported(enum tv_out_mode mode, uint32 cs, int16 width, int16 height) {
	LOG("is_mode_supported\n");
	if (cs == B_CMAP8) return false;
	// B_RGB15, B_RGB16 and B_RGB32 are supported 
	for (int i = 0; supported_resolution[i].width != -1; i++) {
		if (supported_resolution[i].width == width &&
		    supported_resolution[i].height == height) return true;
	}
	if (tv_standard == NTSC && width == 720 && height == 480) return true;
	if (tv_standard == PAL  && width == 720 && height == 576) return true; 	
	LOG("mode is not supported\n");
	return false;
}


bool compare_mode(tv_mode_t *t, display_mode* dm) {
	return t->tv_standard == tv_standard
		&& t->width == dm->timing.h_display
		&& t->height == dm->timing.v_display;
}

bool set_mode(display_mode *dm, enum tv_out_mode mode, set_display_mode_t *set_display_mode) {
	LOG("set_mode\n");
	// remember display mode
	vga_mode = *dm;
	unsigned char hoffset = 2;              // 2 x 4 pixels: TNT2 or TNT2 M64 
	if (riva_version == TNT1) hoffset = 0;  // TNT1 or GeForce2

	for (int i = 0; tv_mode[i].width != -1; i++) {
		if (compare_mode(&tv_mode[i], dm)) {
			int monstat;
			tv_mode_t *m = &tv_mode[i];
#ifdef DEBUG
		char buffer[256];
		sprintf(buffer, "%dx%d %d %d %x\n",
			(int)m->width, (int)m->height,
			(int)m->tv_standard,
			(int)m->ioctl_fct, 
			(int)m->setMode);
		LOG(buffer);
#endif	
			// XXX: add error handling
			// XXX: add DVD modes
			
			//fixme: reset BT should be here...
			//(but: beware of the 'locked SDA' syndrome then!)
			KillClock_Blackout_BT (riva_iic_fd);
			monstat = ReadBT_ConnectStat (riva_iic_fd);
			
			// set vga registers
			m->setMode(set_display_mode, *dm);

			if (m->initBT) m->initBT(riva_iic_fd);

			//modify BT Hphase signal to center TV image...
			SetBT_Hphase (riva_iic_fd, riva_version, m->input);

			//setup BT for TVout (filters, DAC_muxes...)
			SetupBTforTVout (riva_iic_fd, 3, monstat, use_ffilter);

			// ...
			LOG("ioctl...\n");
			ioctl(riva_iic_fd, m->ioctl_fct, &riva_version);
			// turn on/off vga 
			LOG("ioctl 2...\n");
			if (mode == DUAL_MODE) {
				ioctl(riva_iic_fd, rivaVgaOn);
			} else { // mode == TV_OUT_MODE
				ioctl(riva_iic_fd, rivaVgaOff);
			}
			LOG("set_mode successful\n");
			return true;
		}
	}
	LOG("set_mode should not reache here!\n");
	// should not reach here! 
	return false;
}

void load_tv_encoder_settings() {
	// intentionally empty
}


