/*
** Copyright 2001, Michael Pfeiffer	
*/

#include "common.h"

typedef struct {
	int16 width;
	int16 height;
	enum tv_standard tv_standard;
	byte mode;
	byte hpos;
	byte vpos;
	byte oflow;
	int  ioctl_fct;
	int (*setMode)(set_display_mode_t *set_display_mode, display_mode cur_mode);
} tv_mode_t;

/* global data */
static byte              ch_bus;
static byte              ch_addr;
static byte              ch_version;
  
static resolution supported_resolution[] = {
	{640, 480},
	{800, 600},
	{-1, -1}
};

static tv_mode_t tv_mode[] = {
	{640, 480, NTSC, 17, 0x30, 0x02, 0x01, rivaNtsc640, setVESA640},
	{800, 600, NTSC, 24, 0x3e, 0xff, 0x00, rivaNtsc800, setVESA800},
	{640, 480, PAL , 19, 0x38, 0x00, 0x00, rivaPal640,  setVESA640},
	{800, 600, PAL , 20, 0x40, 0x00, 0x00, rivaPal800,  setVESA800},
// XXX: add DVD modes:
//	{800, 600, PAL , 20, 0x40, 0x00, 0x00, rivaPal800,  setVESA800},
//	{800, 600, PAL , 20, 0x40, 0x00, 0x00, rivaPal800,  setVESA800},
	{-1,   -1, PAL ,  0,    0,    0,    0, rivaPal800,  NULL}
};


bool detect_tv_encoder(int fd, int bus_found[2]) {
	LOG("detect_chrontel entered\n");
	for (int i = 0; i < 2; i++) {
		if (bus_found[i]) {
			ch_bus = i;
			if (ch_detect(fd, ch_bus, &ch_addr, &ch_version)) {
				return true;
			}
		}
	}
	LOG("detect_chrontel failed\n");
	return false;
}

void switch_to_vga(set_display_mode_t *set_display_mode) {
	LOG("switch_to_vga\n");
	ch_rud_KillClock_CH(riva_iic_fd, ch_bus, ch_addr);
	LOG("  ioctl rivaVgaOnly\n");
	ioctl(riva_iic_fd, rivaVgaOnly);
	LOG("  set_display_mode\n");
	set_display_mode(&vga_mode);
	LOG("  ch_rud_Reset_CH\n");
	ch_rud_Reset_CH(riva_iic_fd, ch_bus, ch_addr);
	LOG("switch_to_vga done\n");
}

bool is_mode_supported(enum tv_out_mode mode, uint32 cs, int16 width, int16 height) {
	LOG("is_mode_supported\n");
	if (cs == B_CMAP8) return false;
	// B_RGB15, B_RGB16 and B_RGB32 are supported 
	for (int i = 0; supported_resolution[i].width != -1; i++) {
		if (supported_resolution[i].width == width &&
		    supported_resolution[i].height == height) return true;
	}
	if (tv_standard == NTSC && width == 720 && height == 480) return true;
	if (tv_standard == PAL  && width == 720 && height == 576) return true; 	
	LOG("mode is not supported\n");
	return false;
}


bool compare_mode(tv_mode_t *t, display_mode* dm) {
	return t->tv_standard == tv_standard
		&& t->width == dm->timing.h_display
		&& t->height == dm->timing.v_display;
}

bool set_mode(display_mode *dm, enum tv_out_mode mode, set_display_mode_t *set_display_mode) {
	LOG("set_mode\n");
	// remember display mode
	vga_mode = *dm;
	unsigned char hoffset = 2;              // 2 x 4 pixels: TNT2 or TNT2 M64 
	if (riva_version == TNT1) hoffset = 0;  // TNT1 or GeForce2

	for (int i = 0; tv_mode[i].width != -1; i++) {
		if (compare_mode(&tv_mode[i], dm)) {
			tv_mode_t *m = &tv_mode[i];
#ifdef DEBUG
		char buffer[256];
		sprintf(buffer, "%dx%d %d %d %d %d %d %d %x\n",
			(int)m->width, (int)m->height,
			(int)m->tv_standard,
			(int)m->mode, 
			(int)m->hpos, 
			(int)m->vpos, 
			(int)m->oflow, 
			(int)m->ioctl_fct, 
			(int)m->setMode);
		LOG(buffer);
#endif	
			// set vga registers
			m->setMode(set_display_mode, *dm);
			LOG("ch_rud_set_mode\n");
			// program chrontel
			ch_rud_set_mode(riva_iic_fd, ch_bus, ch_addr, 
							m->mode, m->hpos+hoffset, m->vpos, m->oflow, use_ffilter);
			// ...
			LOG("ioctl...\n");
			ioctl(riva_iic_fd, m->ioctl_fct, &riva_version);
			// turn on/off vga 
			LOG("ioctl 2...\n");
			if (mode == DUAL_MODE) {
				ioctl(riva_iic_fd, rivaVgaOn);
			} else { // mode == TV_OUT_MODE
				ioctl(riva_iic_fd, rivaVgaOff);
			}
			LOG("set_mode successful\n");
			return true;
		}
	}
	LOG("set_mode should not reache here!\n");
	// should not reach here! 
	return false;
}

void load_tv_encoder_settings() {
	// intentionally empty
}



