/* riva_iic.c IIC DRIVER V0.13 10/2001
** written by Rudolf Cornelissen
** Email: rag.cornelissen@inter.nl.net
*/

//keep 'X86/_KERNEL_' added in project!!

#include <KernelExport.h>
#include <Drivers.h>
#include <ISA.h>

//#include <module.h>
//#include <Errors.h>
//#include <BeBuild.h>
//#include <stdio.h>
//#include <fcntl.h>

#include "riva_iic.h"

//version control: following line just should be here..
int32 api_version = B_CUR_DRIVER_API_VERSION;

//used NVIDIA TNT/TNT2/GFORCE/GFORCE2 registers: (are fixed at these I/O adresses.)
const unsigned int 	AdresReg = 0x03d4,
					DataReg = 0x03d5;

//used index adresses within NVIDIA chipset:
const unsigned char RDIICBus0 = 0x3e,
					WRIICBus0 = 0x3f,
					RDIICBus1 = 0x36,
					WRIICBus1 = 0x37;

/*
** Bit definition of RDIICBusX: (do not write here!)
** bit7: unknown
** bit6: unknown
** bit5: unknown
** bit4: unknown
** bit3: SDA line input
** bit2: SCL line input
** bit1: unknown
** bit0: unknown
*/

/*
** Bit definition of WRIICBusX:
** bit7: don't change
** bit6: don't change
** bit5: SCL line output (R/W)
** bit4: SDA line output (R/W)
** bit3: write '0'
** bit2: write '0'
** bit1: write '0'
** bit0: write '1'
*/

static spinlock lock = 0;
#define ENTER_CRITICAL_SECTION { cpu_status former; former = disable_interrupts(); acquire_spinlock(&lock);
#define LEAVE_CRITICAL_SECTION release_spinlock(&lock); restore_interrupts(former); }

isa_module_info* ISA_Bus;

status_t
init_hardware (void)
{
	dprintf("riva_iic: driver hardware init called.\n");
	return(B_OK);
}


status_t
init_driver (void)
{
	status_t ret;
  	ret = get_module (B_ISA_MODULE_NAME,(module_info**)&ISA_Bus);
	if (ret<0)
	{
		dprintf ("\nriva_iic: driver init returns negative\n");
		return (ENODEV);
	}
	else
	{
		dprintf("\nriva_iic: driver init returns OK\n");
		
		bstop(0);//init IIC bus 0
		bstop(1);//init IIC bus 1
		
		return(B_OK);
	}
}

void
uninit_driver (void)
{
	put_module(B_ISA_MODULE_NAME);
	dprintf("riva_iic: driver uninited\n");
}	

	
static status_t
iic_device_open (const char *name, uint32 flags, void** cookie)
{
	dprintf("riva_iic: device opened\n");
	return(B_OK);
}


static status_t
iic_device_read (void* cookie, off_t position, void *buffer, size_t* num_bytes)
{
	int bytes = *num_bytes;
	static unsigned char rxdata[40],
						 txack[40];
	int count, BusNR;

//	unsigned char temp;

	memcpy(txack,((char*)buffer),bytes+1);					
	BusNR = txack[0];
	rxdata[0] = BusNR;

	dprintf("riva_iic: reading: %d byte(s) on IIC bus %d\n",bytes,BusNR);

	for (count = 0; count < bytes; count++)
		rxdata[count+1] = RXByte (txack[count+1],BusNR);

	dprintf("riva_iic: data is:\n");
	for (count = 0; count < bytes; count++)
		dprintf("%d -> 0x%x\n", count, (int)rxdata[count+1]);
	dprintf(".\n");
	memcpy(((char*)buffer),rxdata,bytes+1);

/*
			//MEMADR $0068050c:
			//RMA:
    		ISA_Bus->write_io_8(AdresReg,0x38);
    		ISA_Bus->write_io_8(DataReg,0x03);
			//MEMADR:
    		ISA_Bus->write_io_8(0x03d2,0x68);
    		ISA_Bus->write_io_8(0x03d3,0x00);
    		ISA_Bus->write_io_8(0x03d0,0x00);
    		ISA_Bus->write_io_8(0x03d1,0x07);
    		//RMA:
    		ISA_Bus->write_io_8(AdresReg,0x38);
    		ISA_Bus->write_io_8(DataReg,0x05);
			//MEMADR:
//    		longtemp =
    		rxdata[3] = ISA_Bus->read_io_8(0x03d2);// * 0x10000 +
    		rxdata[4] = ISA_Bus->read_io_8(0x03d3);// * 0x1000000;
    		rxdata[1] = ISA_Bus->read_io_8(0x03d0);// +
    		rxdata[2] = ISA_Bus->read_io_8(0x03d1);// * 0x100 +
*/			//unlock VGA registers...
/*    		ISA_Bus->write_io_8(AdresReg,0x11);
			temp = (ISA_Bus->read_io_8(DataReg) & 0x7f);
    		ISA_Bus->write_io_8(AdresReg,0x11);
			ISA_Bus->write_io_8(DataReg,temp);
			
			//SEQ:
    		ISA_Bus->write_io_8(0x03c4,0x01);
    		rxdata[1] = ISA_Bus->read_io_8(0x3c5);

			//hor display end:
    		ISA_Bus->write_io_8(AdresReg,0x01);
    		rxdata[2] = ISA_Bus->read_io_8(DataReg);

			//hor restrace start:
    		ISA_Bus->write_io_8(AdresReg,0x04);
    		rxdata[3] = ISA_Bus->read_io_8(DataReg);

			//hor display total:
    		ISA_Bus->write_io_8(AdresReg,0x00);
    		rxdata[4] = ISA_Bus->read_io_8(DataReg);

			//hor retrace end:
    		ISA_Bus->write_io_8(AdresReg,0x05);
    		rxdata[5] = ISA_Bus->read_io_8(DataReg);

			//hor blank start:
    		ISA_Bus->write_io_8(AdresReg,0x02);
    		rxdata[6] = ISA_Bus->read_io_8(DataReg);

			//hor blank end:
    		ISA_Bus->write_io_8(AdresReg,0x03);
    		rxdata[7] = ISA_Bus->read_io_8(DataReg);

			//LSR:
    		ISA_Bus->write_io_8(AdresReg,0x25);
    		rxdata[8] = ISA_Bus->read_io_8(DataReg);

			//PIXEL:
    		ISA_Bus->write_io_8(AdresReg,0x28);
    		rxdata[9] = ISA_Bus->read_io_8(DataReg);

			//vert retrace start:
    		ISA_Bus->write_io_8(AdresReg,0x10);
    		rxdata[10] = ISA_Bus->read_io_8(DataReg);

			//overflow:
    		ISA_Bus->write_io_8(AdresReg,0x07);
    		rxdata[11] = ISA_Bus->read_io_8(DataReg);

			//vert. display total:
    		ISA_Bus->write_io_8(AdresReg,0x06);
    		rxdata[12] = ISA_Bus->read_io_8(DataReg);

			//vert. retrace end:
    		ISA_Bus->write_io_8(AdresReg,0x11);
    		rxdata[13] = ISA_Bus->read_io_8(DataReg);

			//vert. blank start:
    		ISA_Bus->write_io_8(AdresReg,0x15);
    		rxdata[14] = ISA_Bus->read_io_8(DataReg);

			//height (vert. blank start b9):
    		ISA_Bus->write_io_8(AdresReg,0x09);
    		rxdata[15] = ISA_Bus->read_io_8(DataReg);

			//vert. blank end:
    		ISA_Bus->write_io_8(AdresReg,0x16);
    		rxdata[16] = ISA_Bus->read_io_8(DataReg);

			//vert. disp end:
    		ISA_Bus->write_io_8(AdresReg,0x12);
    		rxdata[17] = ISA_Bus->read_io_8(DataReg);

			//misc. output:
    		rxdata[18] = ISA_Bus->read_io_8(0x03cc);
    		
			//REPAINT1:
    		ISA_Bus->write_io_8(AdresReg,0x1a);
 			rxdata[19] = ISA_Bus->read_io_8(DataReg);


	rxdata[0] = 0xff;//test
	memcpy(((char*)buffer),rxdata,20);
*/
	return (B_OK);
}


static status_t
iic_device_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes)
{
	int bytes = *num_bytes;
	static unsigned char data[40];
	int count, BusNR;
	
	memcpy(data,((char*)buffer),bytes+1);					
	BusNR = data[0];
	
	dprintf("riva_iic: writing: %d byte(s) on IIC bus %d\nriva_iic: data is:",bytes,BusNR);
	for (count = 0; count < bytes; count++)
		dprintf("%d -> 0x%x\n", count, (int)data[count+1]);
	dprintf(".\n");

	for (count = 0; count < bytes; count++)
		TXByte (data[count+1],BusNR);
	
	return (B_OK);
}


static status_t
iic_device_control (void* cookie, uint32 operation, void* args, size_t length)
{
	int err = B_OK;
	int BusNR;
	unsigned char lsb, msb, type,
					 temp, temp1, temp2, temp3;//for VESA VGA switchback (see below)...
	static unsigned char data[1];

	memcpy(data,((char*)args),1);					
	BusNR = data[0];	//for IIC operations
	type = data[0];		//for RIVA operations

	dprintf("riva_iic: device control (%d/0x%x) called\n", (int)operation, (int)operation);
 
    switch (operation)
    { 
    case rivaStartIIC:	
    		dprintf("riva_iic: generate start condition on IIC bus %d\n",BusNR);  
			FlagIICError (-1);//clear IIC errorcode
			bstart(BusNR);
			if (FlagIICError (0)) err = B_IO_ERROR;
            break; 
	case rivaStopIIC:
			dprintf("riva_iic: generate stop condition on IIC bus %d\n",BusNR);
			bstop(BusNR);
			if (FlagIICError (0)) err = B_IO_ERROR;
			//return IIC err code here in the future...
			FlagIICError (-1);//clear IIC errorcode
			break;
	case rivaChkBus:
			snooze(5);
			bstop(BusNR);//init and/or stop comms IIC bus
			dprintf("riva_iic: check for existance of IIC bus %d: ",BusNR);
			//check for hardware coupling of SCL and SDA -out and -in lines.
			snooze(5);
			OutSCL(0,BusNR);
			OutSDA(1,BusNR);
			snooze(3);
			if (InSCL(BusNR) || !InSDA(BusNR))
			{
				err = B_IO_ERROR;
				dprintf("fail.\n");
				break;
			}
			snooze(2);
			OutSCL(1,BusNR);
			OutSDA(0,BusNR);
			snooze(3);
			if (!InSCL(BusNR) || InSDA(BusNR))
			{
				err = B_IO_ERROR;
				dprintf("fail.\n");
				break;
			}
			dprintf("OK.\n");
			snooze(2);
			bstop(BusNR);//re-init IIC bus
			break;
	case rivaVersion:
			data[0] = 13; //version number = 13/100 = 0.13
			dprintf("riva_iic: return driver version nr.: %d/100\n",data[0]);
			memcpy(((char*)args),data,1);					
			break;
    case rivaUnlock:
    		dprintf("riva_iic: unlocking extended registers for R/W\n");
			ENTER_CRITICAL_SECTION
    		ISA_Bus->write_io_8(AdresReg,0x1f);
 			ISA_Bus->write_io_8(DataReg,0x57);//write 'magic number'
 			LEAVE_CRITICAL_SECTION
    		break;
    case rivaPal640:
    		dprintf("riva_iic: switch RIVA chip for (dualhead clone) PAL 50Hz 640x480\n");
			SwitchRIVAtoTV (0x39, 0x02, type);
	   		break;
    case rivaPal800:
    		dprintf("riva_iic: switch RIVA chip for (dualhead clone) PAL 50Hz 800x600\n");
    		
			SwitchRIVAtoTV (0xb6, 0x02, type);
	   		break;
    case rivaNtsc640:
    		dprintf("riva_iic: switch RIVA chip for (dualhead clone) NTSC 60Hz 640x480\n");
			SwitchRIVAtoTV (0x2a, 0x02, type);
	   		break;
    case rivaNtsc800:
    		dprintf("riva_iic: switch RIVA chip for (dualhead clone) NTSC 60Hz 800x600\n");
			SwitchRIVAtoTV (0xac, 0x02, type);
	   		break;
	case rivaEnableTVOut:
			dprintf("riva_iic: switch RIVA chip for tv out\n");
			EnableTVOut();
			break;
	case rivaVgaOff:
			dprintf("riva_iic: turn off VGA monitor\n");
			//REPAINT1: switch off VGA monitor HSYNC
			ENTER_CRITICAL_SECTION
    		ISA_Bus->write_io_8(AdresReg,0x1a);
 			ISA_Bus->write_io_8(DataReg,(ISA_Bus->read_io_8(DataReg) | 0x80));
 			LEAVE_CRITICAL_SECTION
	   		break;
	case rivaVgaOn:
			dprintf("riva_iic: turn on VGA monitor\n");
			//REPAINT1: switch on VGA monitor HSYNC
			ENTER_CRITICAL_SECTION
    		ISA_Bus->write_io_8(AdresReg,0x1a);
 			ISA_Bus->write_io_8(DataReg,(ISA_Bus->read_io_8(DataReg) & 0x7f));
 			LEAVE_CRITICAL_SECTION
	   		break;
	case rivaVesaVgaOnly:
			dprintf("riva_iic: switch RIVA chip for (VESA) VGA only\n");
			//tested on ELSA Erazor III 32Mb AGP (TNT2/BT869)
			//and Diamond Viper V550 16Mb PCI (TNT1/BT869)

			//works OK if booted in VESA on VGA or TV;
			//(sets refreshrate = 60Hz).
			//(for non VESA VGA the BScreen API and 'rivaVGAOnly' should be used.)

			//switch RIVA to VGA:
			SwitchToVGA();

			//set RIVA picture timing registers (VESA setup used):
			//read hor display end:
			ENTER_CRITICAL_SECTION
    		ISA_Bus->write_io_8(AdresReg,0x01);
			if ((ISA_Bus->read_io_8(DataReg)) == 0x63)
			{
				//800x600 resolution active...
				lsb = 0x72;//values for vertical total...
				msb = 0x02;

				//values for RIVA pixelclock (60Hz):
 				temp = 0x0b;
			    temp1 = 0x7b;
			    temp2 = 0x02;
			    temp3 = 0x00;

				//set hor blank start:
    			ISA_Bus->write_io_8(AdresReg,0x02);
    			ISA_Bus->write_io_8(DataReg,0x63);

				//set hor restrace start:
    			ISA_Bus->write_io_8(AdresReg,0x04);
    			ISA_Bus->write_io_8(DataReg,0x6a);

				//set hor display total:
    			ISA_Bus->write_io_8(AdresReg,0x00);
    			ISA_Bus->write_io_8(DataReg,0x7f);

				//set hor retrace end:
    			ISA_Bus->write_io_8(AdresReg,0x05);
				ISA_Bus->write_io_8(DataReg,0x1a);

				//set hor blank end:
    			ISA_Bus->write_io_8(AdresReg,0x03);
				ISA_Bus->write_io_8(DataReg,0x83);

				//set LSR:
    			ISA_Bus->write_io_8(AdresReg,0x25);
				ISA_Bus->write_io_8(DataReg,0x00);

				//set vertical display end:
    			ISA_Bus->write_io_8(AdresReg,0x12);
				ISA_Bus->write_io_8(DataReg,0x57);

				//set vert blank start:
    			ISA_Bus->write_io_8(AdresReg,0x15);
    			ISA_Bus->write_io_8(DataReg,0x57);

				//set overflow:
    			ISA_Bus->write_io_8(AdresReg,0x07);
				ISA_Bus->write_io_8(DataReg,0xf0);

				//set height:
		  		ISA_Bus->write_io_8(AdresReg,0x09);
				ISA_Bus->write_io_8(DataReg,0x60);

				//set vert retrace start:
    			ISA_Bus->write_io_8(AdresReg,0x10);
    			ISA_Bus->write_io_8(DataReg,0x59);

				//set vert. retrace end:
    			ISA_Bus->write_io_8(AdresReg,0x11);
				ISA_Bus->write_io_8(DataReg,0x0d);

				//set vert. blank end:
    			ISA_Bus->write_io_8(AdresReg,0x16);
    			ISA_Bus->write_io_8(DataReg,0x73);
			}
			else
			{
				//assume 640x480 resolution active...
				lsb = 0x0b;//values for vertical total...
				msb = 0x02;

				//values for RIVA pixelclock (60Hz):
			    temp = 0x0e;
			    temp1 = 0xc5;
			    temp2 = 0x03;
			    temp3 = 0x00;

				//set hor blank start:
	    		ISA_Bus->write_io_8(AdresReg,0x02);
    			ISA_Bus->write_io_8(DataReg,0x4f);

				//set hor restrace start:
    			ISA_Bus->write_io_8(AdresReg,0x04);
    			ISA_Bus->write_io_8(DataReg,0x53);

				//set hor display total:
    			ISA_Bus->write_io_8(AdresReg,0x00);
    			ISA_Bus->write_io_8(DataReg,0x5f);

				//set hor retrace end:
    			ISA_Bus->write_io_8(AdresReg,0x05);
				ISA_Bus->write_io_8(DataReg,0x9f);

				//set hor blank end:
    			ISA_Bus->write_io_8(AdresReg,0x03);
				ISA_Bus->write_io_8(DataReg,0x83);

				//set LSR:
    			ISA_Bus->write_io_8(AdresReg,0x25);
				ISA_Bus->write_io_8(DataReg,0x00);

				//set vertical display end:
    			ISA_Bus->write_io_8(AdresReg,0x12);
				ISA_Bus->write_io_8(DataReg,0xdf);

				//set vert blank start:
    			ISA_Bus->write_io_8(AdresReg,0x15);
    			ISA_Bus->write_io_8(DataReg,0xdf);

				//set overflow:
    			ISA_Bus->write_io_8(AdresReg,0x07);
				ISA_Bus->write_io_8(DataReg,0x3e);

				//set height:
		  		ISA_Bus->write_io_8(AdresReg,0x09);
				ISA_Bus->write_io_8(DataReg,0x40);

				//set vert retrace start:
    			ISA_Bus->write_io_8(AdresReg,0x10);
    			ISA_Bus->write_io_8(DataReg,0xea);

				//set vert. retrace end:
    			ISA_Bus->write_io_8(AdresReg,0x11);
				ISA_Bus->write_io_8(DataReg,0x0c);

				//set vert. blank end:
    			ISA_Bus->write_io_8(AdresReg,0x16);
    			ISA_Bus->write_io_8(DataReg,0x0c);
			}

			//Set the vertical total:
    		//LSB (vertical total):
    		ISA_Bus->write_io_8(AdresReg,0x06);
    		ISA_Bus->write_io_8(DataReg,lsb);
			//b8 and b9:(overflow)
    		ISA_Bus->write_io_8(AdresReg,0x07);
			ISA_Bus->write_io_8(DataReg,
				((ISA_Bus->read_io_8(DataReg) & ~0x21) |
					((msb & 0x01)+((msb & 0x02)<<4))));
			//b10:(LSR)
	  		ISA_Bus->write_io_8(AdresReg,0x25);
			ISA_Bus->write_io_8(DataReg,
				((ISA_Bus->read_io_8(DataReg) & ~0x01) | ((msb & 0x04)>>2)));


			//set RIVA pixelclock:
			//(MEMADR $00680508)

			//write adress:
			//RMA:
		    ISA_Bus->write_io_8(AdresReg,0x38);
		    ISA_Bus->write_io_8(DataReg,0x03);
			//MEMADR:
    		ISA_Bus->write_io_8(0x03d0,0x08);
    		ISA_Bus->write_io_8(0x03d1,0x05);
    		ISA_Bus->write_io_8(0x03d2,0x68);
    		ISA_Bus->write_io_8(0x03d3,0x00);
    		//RMA:
    		ISA_Bus->write_io_8(AdresReg,0x38);
    		ISA_Bus->write_io_8(DataReg,0x07);
			//MEMADR:
    		ISA_Bus->write_io_8(0x03d0,temp);
    		ISA_Bus->write_io_8(0x03d1,temp1);
    		ISA_Bus->write_io_8(0x03d2,temp2);
    		ISA_Bus->write_io_8(0x03d3,temp3);
    		//RMA:(not nessesary?)
    		ISA_Bus->write_io_8(AdresReg,0x38);
    		ISA_Bus->write_io_8(DataReg,0x03);
			LEAVE_CRITICAL_SECTION
	   		break;
	case rivaVgaOnly:
			dprintf("riva_iic: switch RIVA chip for (non-VESA) VGA only\n");
			//switch RIVA to VGA:
			SwitchToVGA();
	   		break;
	case rivaIsMode640:
			dprintf("riva_iic: IsMode 640x480 check\n");
			ENTER_CRITICAL_SECTION
			//unlock VGA registers at index 0-7
    		ISA_Bus->write_io_8(AdresReg,0x11);
			ISA_Bus->write_io_8(DataReg,(ISA_Bus->read_io_8(DataReg) & 0x7f));
			//check hor display end for 640 pixels
    		ISA_Bus->write_io_8(AdresReg,0x01);
			if (!((ISA_Bus->read_io_8(DataReg)) == 0x4f)) err = B_IO_ERROR;
			LEAVE_CRITICAL_SECTION
			break;
	case rivaIsModeValid:
			dprintf("riva_iic: IsModeValid check\n");
			ENTER_CRITICAL_SECTION
			//unlock VGA registers at index 0-7
    		ISA_Bus->write_io_8(AdresReg,0x11);
			ISA_Bus->write_io_8(DataReg,(ISA_Bus->read_io_8(DataReg) & 0x7f));
			//check hor display end for 640 or 800 pixels
    		ISA_Bus->write_io_8(AdresReg,0x01);
			if (!((ISA_Bus->read_io_8(DataReg)) == 0x4f) &&
					!((ISA_Bus->read_io_8(DataReg)) == 0x63)) err = B_IO_ERROR;
			LEAVE_CRITICAL_SECTION
			break;
    default:dprintf("riva_iic: illegal ioctrl command\n");
            err = B_DEV_INVALID_IOCTL;
            break;
    }        
    
	dprintf("leave iic_device_control\n");
	return(err);
}


static status_t
iic_device_close (void* cookie)
{
	dprintf("riva_iic: device closed\n");
	return B_OK;
}


static status_t
iic_device_free (void* cookie)
{
	return B_OK;
}


static const char *my_device_name[] = 
{
	"video/riva_iic/1",
	NULL
};


device_hooks my_device_hooks = 
{
	iic_device_open, 			/* -> open entry point */
	iic_device_close, 			/* -> close entry point */
	iic_device_free,			/* -> free cookie */
	iic_device_control, 		/* -> control entry point */
	iic_device_read,			/* -> read entry point */
	iic_device_write			/* -> write entry point */
};


const char**
publish_devices()
{
	return my_device_name;
}


device_hooks*
find_device(const char* name)
{
	return &my_device_hooks;
}

void SwitchToVGA()
{
	//tested on ELSA Erazor III 32Mb AGP (TNT2/BT869)
	//and Diamond Viper V550 16Mb PCI (TNT1/BT869)

	unsigned char temp, temp1, temp2, temp3;
	unsigned char VertRetrace, ImageCounter;
	
	ENTER_CRITICAL_SECTION
	//REPAINT1: switch on VGA monitor HSYNC and VSYNC
	ISA_Bus->write_io_8(AdresReg,0x1a);
	ISA_Bus->write_io_8(DataReg,(ISA_Bus->read_io_8(DataReg) & 0x3f));

	//SEQ:
	ISA_Bus->write_io_8(0x03c4,0x01);
	ISA_Bus->write_io_8(0x03c5,0x01);

	//select 'attribute control' VGA register $20
	//(select index adress register:)
	ISA_Bus->read_io_8(0x3da);
	//(write index for register $20:)
	ISA_Bus->write_io_8(0x03c0,0x20);
	//(select index adress register again:)
	ISA_Bus->read_io_8(0x3da);

	//wait for 50 images (about 0.5-1.0 second) to be generated to make sure
	// VGA has kicked in and is running OK before continuing...
	VertRetrace = 1;
	ImageCounter = 1; // XXX: read from settings file:
	//I suggest we leave ImageCounter at '1' after all; until proven wrong: This
	//will be the case if switchback doesn't always function in VESA mode. Rudolf.
	do
	{
		//(make sure we are 'in' active VGA picture:)
		do  VertRetrace = (ISA_Bus->read_io_8(0x03da) & 0x08);
			while (VertRetrace);
		//(wait for vertical retrace start on VGA:)
		do  VertRetrace = (ISA_Bus->read_io_8(0x03da) & 0x08);
			while (!VertRetrace);
		//(make sure we are 'in' active VGA picture again:)
		do  VertRetrace = (ISA_Bus->read_io_8(0x03da) & 0x08);
			while (VertRetrace);
		//decrement picture counter:
		ImageCounter --;
	} while (ImageCounter);

	//'update' PIXEL/TV:
	ISA_Bus->write_io_8(AdresReg,0x28);
	ISA_Bus->write_io_8(DataReg,(ISA_Bus->read_io_8(DataReg) & 0x03));

	//VGA reg $21... (enable reprogram)
	ISA_Bus->write_io_8(AdresReg,0x21);
	ISA_Bus->write_io_8(DataReg,0xfa);

	//CAUTION:
	//On the TNT1, these memadresses apparantly cannot be read (sometimes)!;
	//write actions do succeed though...
	//($00680700 b1-23 and b25-31 apparantly are 'don't cares'...)

	//SWITCH RIVA pixelclock to be RIVA's own (so: directly):
	//MEMADR $00680700:
	//read adress:
	//RMA:
	ISA_Bus->write_io_8(AdresReg,0x38);
	ISA_Bus->write_io_8(DataReg,0x03);
	//MEMADR:
	ISA_Bus->write_io_8(0x03d0,0x00);
	ISA_Bus->write_io_8(0x03d1,0x07);
	ISA_Bus->write_io_8(0x03d2,0x68);
	ISA_Bus->write_io_8(0x03d3,0x00);
	//RMA:
	ISA_Bus->write_io_8(AdresReg,0x38);
	ISA_Bus->write_io_8(DataReg,0x05);
	//MEMADR:
//  longtemp =
    temp = ISA_Bus->read_io_8(0x03d0);// +
    temp1 = ISA_Bus->read_io_8(0x03d1);// * 0x100 +
    temp2 = ISA_Bus->read_io_8(0x03d2);// * 0x10000 +
    temp3 = ISA_Bus->read_io_8(0x03d3);// * 0x1000000;

	//modify read data:
	temp = temp & ~0x01;
	//and make sure b24 is always '1':
	temp3 = temp3 | 0x01;

	//write adress:
	//RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,0x00);
    ISA_Bus->write_io_8(0x03d1,0x07);
    ISA_Bus->write_io_8(0x03d2,0x68);
    ISA_Bus->write_io_8(0x03d3,0x00);
    //RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x07);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,temp);
    ISA_Bus->write_io_8(0x03d1,temp1);
    ISA_Bus->write_io_8(0x03d2,temp2);
    ISA_Bus->write_io_8(0x03d3,temp3);
    //RMA:(not nessesary?)
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);

	//switch RIVA PLL phase-lock to lock to RIVA's own internal pixelclock:
	//MEMADR $0068050c:
	//create data:
    temp = 0x00;
    temp1 = 0x07;
    temp2 = 0x00;
    temp3 = 0x10;

	//write adress:
	//RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,0x0c);
    ISA_Bus->write_io_8(0x03d1,0x05);
    ISA_Bus->write_io_8(0x03d2,0x68);
    ISA_Bus->write_io_8(0x03d3,0x00);
    //RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x07);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,temp);
    ISA_Bus->write_io_8(0x03d1,temp1);
    ISA_Bus->write_io_8(0x03d2,temp2);
    ISA_Bus->write_io_8(0x03d3,temp3);
    //RMA:(not nessesary?)
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);

	//TREG:
	ISA_Bus->write_io_8(AdresReg,0x3d);
	ISA_Bus->write_io_8(DataReg,0x00);


	//MEMADR $00680880:
	//read adress:
	//RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,0x80);
    ISA_Bus->write_io_8(0x03d1,0x08);
    ISA_Bus->write_io_8(0x03d2,0x68);
    ISA_Bus->write_io_8(0x03d3,0x00);
    //RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x05);
	//MEMADR:
//  longtemp =
    temp = ISA_Bus->read_io_8(0x03d0);// +
    temp1 = ISA_Bus->read_io_8(0x03d1);// * 0x100 +
    temp2 = ISA_Bus->read_io_8(0x03d2);// * 0x10000 +
    temp3 = ISA_Bus->read_io_8(0x03d3);// * 0x1000000;

	//modify read data:
	temp3 = temp3 | 0x10;

	//write adress:
	//RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,0x80);
    ISA_Bus->write_io_8(0x03d1,0x08);
    ISA_Bus->write_io_8(0x03d2,0x68);
    ISA_Bus->write_io_8(0x03d3,0x00);
    //RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x07);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,temp);
    ISA_Bus->write_io_8(0x03d1,temp1);
    ISA_Bus->write_io_8(0x03d2,temp2);
    ISA_Bus->write_io_8(0x03d3,temp3);
    //RMA:(not nessesary?)
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);

   	//LCD:
    ISA_Bus->write_io_8(AdresReg,0x33);
 	ISA_Bus->write_io_8(DataReg,(ISA_Bus->read_io_8(DataReg) & 0xfc));


	//Set overscan color to 'black':
	//Disable this part if you're trying to center the output on TV,
	//you'll get blue overscan range color then. Use this as a guide-'line' ;-)
	//(select index adress register:)
	ISA_Bus->read_io_8(0x3da);
	//(write index for 'overscan color' register:)
  	ISA_Bus->write_io_8(0x03c0,0x11);
    //(write data for 'overscan color' register:)
    ISA_Bus->write_io_8(0x03c0,0x00);


	//unlock VGA registers at index 0-7
    ISA_Bus->write_io_8(AdresReg,0x11);
	temp = (ISA_Bus->read_io_8(DataReg) & 0x7f);
	ISA_Bus->write_io_8(DataReg,temp);

	//unlock VGA registers 'vert. retrace start' & 'vert. retrace end':
    ISA_Bus->write_io_8(AdresReg,0x03);
	ISA_Bus->write_io_8(DataReg,
		(ISA_Bus->read_io_8(DataReg) | 0x80));
	LEAVE_CRITICAL_SECTION
}//end SwitchToVGA.

void SwitchRIVAtoTV (char lsb, char msb, char type)
{
	unsigned char temp, temp1, temp2, temp3;

	//tested on ELSA Erazor III 32Mb AGP (TNT2/BT869);
	//Diamond Viper V550 16Mb PCI (TNT1/BT869);
	//and ASUS V7100 GeForce2 MX200 AGP/32Mb (CH7007)

	//CAUTION:
	//On the TNT1, these memadresses apparantly cannot be read (sometimes)!;
	//write actions do succeed though...
	//($00680700 b1-23 and b25-31 apparantly are 'don't cares'...)

	ENTER_CRITICAL_SECTION
	//SWITCH RIVA pixelclock to be 'via' TVout chip:
	//MEMADR $00680700:
	//read adress:
	//RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,0x00);
    ISA_Bus->write_io_8(0x03d1,0x07);
    ISA_Bus->write_io_8(0x03d2,0x68);
    ISA_Bus->write_io_8(0x03d3,0x00);
    //RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x05);
	//MEMADR:
//  longtemp =
    temp = ISA_Bus->read_io_8(0x03d0);// +
    temp1 = ISA_Bus->read_io_8(0x03d1);// * 0x100 +
    temp2 = ISA_Bus->read_io_8(0x03d2);// * 0x10000 +
    temp3 = ISA_Bus->read_io_8(0x03d3);// * 0x1000000;

	//modify read data:
	temp = temp | 0x01;
	//and make sure b24 is always '1':
	temp3 = temp3 | 0x01;

	//write adress:
	//RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,0x00);
    ISA_Bus->write_io_8(0x03d1,0x07);
    ISA_Bus->write_io_8(0x03d2,0x68);
    ISA_Bus->write_io_8(0x03d3,0x00);
    //RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x07);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,temp);
    ISA_Bus->write_io_8(0x03d1,temp1);
    ISA_Bus->write_io_8(0x03d2,temp2);
    ISA_Bus->write_io_8(0x03d3,temp3);
    //RMA:(not nessesary?)
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);

	//switch RIVA PLL phase-lock to lock to pixelclock from TV-chip instead of
	//internal pixelclock:
	//(nessecary or display will 'shiver' on both TV and VGA.)
	//MEMADR $0068050c:
	//create data:
    temp = 0x00;
    temp1 = 0x07;
    temp2 = 0x03;
    temp3 = 0x00;

	//write adress:
	//RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,0x0c);
    ISA_Bus->write_io_8(0x03d1,0x05);
    ISA_Bus->write_io_8(0x03d2,0x68);
    ISA_Bus->write_io_8(0x03d3,0x00);
    //RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x07);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,temp);
    ISA_Bus->write_io_8(0x03d1,temp1);
    ISA_Bus->write_io_8(0x03d2,temp2);
    ISA_Bus->write_io_8(0x03d3,temp3);
    //RMA:(not nessesary?)
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);


	//VGA reg $21... (enable reprogram!)
    ISA_Bus->write_io_8(AdresReg,0x21);
 	ISA_Bus->write_io_8(DataReg,0xfa);

	//SEQ:
    ISA_Bus->write_io_8(0x03c4,0x01);
    ISA_Bus->write_io_8(0x03c5,0x01);


	//Set overscan color to 'black':
	//Disable this part if you're trying to center the output on TV,
	//you'll get blue overscan range color then. Use this as a guide-'line' ;-)
	//(select index adress register:)
	ISA_Bus->read_io_8(0x3da);
	//(write index for 'overscan color' register:)
    ISA_Bus->write_io_8(0x03c0,0x11);
    //(write data for 'overscan color' register:)
    ISA_Bus->write_io_8(0x03c0,0x00);


	//unlock VGA registers at index 0-7
    ISA_Bus->write_io_8(AdresReg,0x11);
	temp = (ISA_Bus->read_io_8(DataReg) & 0x7f);
	ISA_Bus->write_io_8(DataReg,temp);

	//unlock VGA registers 'vert. retrace start' & 'vert. retrace end':
    ISA_Bus->write_io_8(AdresReg,0x03);
	ISA_Bus->write_io_8(DataReg,
		(ISA_Bus->read_io_8(DataReg) | 0x80));

	//execute RIVA picture timing formula:
	//read hor display end:
    ISA_Bus->write_io_8(AdresReg,0x01);
    temp = ISA_Bus->read_io_8(DataReg);

	//copy to hor blank start:
    ISA_Bus->write_io_8(AdresReg,0x02);
    ISA_Bus->write_io_8(DataReg,(temp));

	//add 1 and place in hor restrace start:
    ISA_Bus->write_io_8(AdresReg,0x04);
    ISA_Bus->write_io_8(DataReg,(temp+1));

	if (type == 3)//TNT2 M64...
	{
		//add 2 and copy b4-0 to hor retrace end:
    	ISA_Bus->write_io_8(AdresReg,0x05);
		ISA_Bus->write_io_8(DataReg,
			((ISA_Bus->read_io_8(DataReg) & ~0x1f) | ((temp+2) & 0x1f)));

		//add 3 and place in hor display total:
    	ISA_Bus->write_io_8(AdresReg,0x00);
    	ISA_Bus->write_io_8(DataReg,(temp+3));
    }
    else//TNT1, TNT2, Geforce2...
	{
		//add 2 and place in hor display total:
    	ISA_Bus->write_io_8(AdresReg,0x00);
   		ISA_Bus->write_io_8(DataReg,(temp+2));

		//add 3 and copy b4-0 to hor retrace end:
	    ISA_Bus->write_io_8(AdresReg,0x05);
		ISA_Bus->write_io_8(DataReg,
			((ISA_Bus->read_io_8(DataReg) & ~0x1f) | ((temp+3) & 0x1f)));
	}

	//add 6 and copy b4-0 to hor blank end:
    ISA_Bus->write_io_8(AdresReg,0x03);
	ISA_Bus->write_io_8(DataReg,
		((ISA_Bus->read_io_8(DataReg) & ~0x1f) | ((temp+6) & 0x1f)));

	//copy hor blank end b5 in hor retrace end:
    ISA_Bus->write_io_8(AdresReg,0x05);
	ISA_Bus->write_io_8(DataReg,
		((ISA_Bus->read_io_8(DataReg) & 0x7f)) | (((temp+6) & 0x20)<<2));

	//clear hor blank end b6 in LSR:
    ISA_Bus->write_io_8(AdresReg,0x25);
	ISA_Bus->write_io_8(DataReg,
		(ISA_Bus->read_io_8(DataReg) & 0xef));

  	//'update' PIXEL/TV:
    ISA_Bus->write_io_8(AdresReg,0x28);
	temp = (ISA_Bus->read_io_8(DataReg) & 0xc7);
 	ISA_Bus->write_io_8(DataReg,temp | 0x80);

	//read vertical display end:
    //(LSB:)
    ISA_Bus->write_io_8(AdresReg,0x12);
	temp = ISA_Bus->read_io_8(DataReg);

	//b8 and b9:(overflow)
    ISA_Bus->write_io_8(AdresReg,0x07);
	temp2 = (ISA_Bus->read_io_8(DataReg) & 0x42);

	//b10:(LSR)
    ISA_Bus->write_io_8(AdresReg,0x25);
	temp3 = (ISA_Bus->read_io_8(DataReg) & 0x02);
	//(combine MSB:)
	temp1 = (((temp2 & 0x02)>>1) + ((temp2 & 0x40)>>5) + (temp3<<1));

	//copy to vert blank start:
	//LSB:
    ISA_Bus->write_io_8(AdresReg,0x15);
    ISA_Bus->write_io_8(DataReg,temp);
	//b8:(overflow)
    ISA_Bus->write_io_8(AdresReg,0x07);
	ISA_Bus->write_io_8(DataReg,
		((ISA_Bus->read_io_8(DataReg) & ~0x08) | ((temp1 & 0x01)<<3)));
	//b9:(height)
	ISA_Bus->write_io_8(AdresReg,0x09);
	ISA_Bus->write_io_8(DataReg,
		((ISA_Bus->read_io_8(DataReg) & ~0x20) | ((temp1 & 0x02)<<4)));
	//b10:(LSR)
	ISA_Bus->write_io_8(AdresReg,0x25);
	ISA_Bus->write_io_8(DataReg,
		((ISA_Bus->read_io_8(DataReg) & ~0x08) | ((temp1 & 0x04)<<1)));

	//add 1 and place in vert retrace start:
	temp++;
	if (temp == 0) temp1++;
	//LSB:
    ISA_Bus->write_io_8(AdresReg,0x10);
    ISA_Bus->write_io_8(DataReg,temp);
	//b8 and b9:(overflow)
    ISA_Bus->write_io_8(AdresReg,0x07);
	ISA_Bus->write_io_8(DataReg,
		((ISA_Bus->read_io_8(DataReg) & ~0x84) |
			(((temp1 & 0x01)<<2)+((temp1 & 0x02)<<6))));
	//b10:(LSR)
	ISA_Bus->write_io_8(AdresReg,0x25);
	ISA_Bus->write_io_8(DataReg,
		((ISA_Bus->read_io_8(DataReg) & ~0x04) | (temp1 & 0x04)));

	//place '(vertical total + 1)' b3-0 in vert retrace end:
	//(This takes care of the 'cursor trash' on TNT1's...)
    ISA_Bus->write_io_8(AdresReg,0x11);
	ISA_Bus->write_io_8(DataReg,
		((ISA_Bus->read_io_8(DataReg) & 0xf0) |	((lsb + 1) & 0x0f)));
			
	//place '(vertical total + 1)' LSB in vert. blank end:
	//(This takes care of the 'cursor trash' on TNT1's...)
    ISA_Bus->write_io_8(AdresReg,0x16);
    ISA_Bus->write_io_8(DataReg,(lsb + 1));
	//End RIVA Timing formula.

	//Set the vertical 'refresh rate':
    //LSB (vertical total):
    ISA_Bus->write_io_8(AdresReg,0x06);
    ISA_Bus->write_io_8(DataReg,lsb);
	//b8 and b9:(overflow)
    ISA_Bus->write_io_8(AdresReg,0x07);
	ISA_Bus->write_io_8(DataReg,
		((ISA_Bus->read_io_8(DataReg) & ~0x21) |
			((msb & 0x01)+((msb & 0x02)<<4))));
	//b10:(LSR)
	ISA_Bus->write_io_8(AdresReg,0x25);
	ISA_Bus->write_io_8(DataReg,
		((ISA_Bus->read_io_8(DataReg) & ~0x01) | ((msb & 0x04)>>2)));

  	//LCD:
    ISA_Bus->write_io_8(AdresReg,0x33);
	temp = (ISA_Bus->read_io_8(DataReg) & 0xfe);
    ISA_Bus->write_io_8(AdresReg,0x33);
 	ISA_Bus->write_io_8(DataReg,temp);

	//TREG:
	ISA_Bus->write_io_8(AdresReg,0x3d);
	ISA_Bus->write_io_8(DataReg,0x80);
	LEAVE_CRITICAL_SECTION
}

void EnableTVOut ()
{
	//temporary, unless a better formula 'will be found' than already in riva_iic:
	//the formula used in riva_iic has to stay to keep the possibility open to enable
	//TVout on non-supported (qua videodriver) VGA chipsets in VESA mode.
	unsigned char temp, temp1, temp2, temp3;

	//tested on ELSA Erazor III 32Mb AGP (TNT2/BT869),
	//Diamond Viper V550 16Mb PCI (TNT1/BT869),
	//and ASUS V7100 GeForce2 MX200 AGP/32Mb (CH7007)

	//CAUTION:
	//On the TNT1, these memadresses apparantly cannot be read (sometimes)!;
	//write actions do succeed though...
	//($00680700 b1-23 and b25-31 apparantly are 'don't cares'...)

	ENTER_CRITICAL_SECTION
	//SWITCH RIVA pixelclock to be 'via' TVout chip:
	//MEMADR $00680700:
	//read adress:
	//RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,0x00);
    ISA_Bus->write_io_8(0x03d1,0x07);
    ISA_Bus->write_io_8(0x03d2,0x68);
    ISA_Bus->write_io_8(0x03d3,0x00);
    //RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x05);
	//MEMADR:
//  longtemp =
    temp = ISA_Bus->read_io_8(0x03d0);// +
    temp1 = ISA_Bus->read_io_8(0x03d1);// * 0x100 +
    temp2 = ISA_Bus->read_io_8(0x03d2);// * 0x10000 +
    temp3 = ISA_Bus->read_io_8(0x03d3);// * 0x1000000;

	//modify read data:
	temp = temp | 0x01;
	//and make sure b24 is always '1':
	temp3 = temp3 | 0x01;

	//write adress:
	//RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,0x00);
    ISA_Bus->write_io_8(0x03d1,0x07);
    ISA_Bus->write_io_8(0x03d2,0x68);
    ISA_Bus->write_io_8(0x03d3,0x00);
    //RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x07);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,temp);
    ISA_Bus->write_io_8(0x03d1,temp1);
   	ISA_Bus->write_io_8(0x03d2,temp2);
    ISA_Bus->write_io_8(0x03d3,temp3);
    //RMA:(not nessesary?)
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);

	//switch RIVA PLL phase-lock to lock to pixelclock from TV-chip instead of
	//internal pixelclock:
	//(nessecary or display will 'shiver' on both TV and VGA.)
	//MEMADR $0068050c:
	//create data:
    temp = 0x00;
    temp1 = 0x07;
    temp2 = 0x03;
    temp3 = 0x00;

	//write adress:
	//RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,0x0c);
    ISA_Bus->write_io_8(0x03d1,0x05);
    ISA_Bus->write_io_8(0x03d2,0x68);
    ISA_Bus->write_io_8(0x03d3,0x00);
    //RMA:
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x07);
	//MEMADR:
    ISA_Bus->write_io_8(0x03d0,temp);
    ISA_Bus->write_io_8(0x03d1,temp1);
    ISA_Bus->write_io_8(0x03d2,temp2);
    ISA_Bus->write_io_8(0x03d3,temp3);
    //RMA:(not nessesary?)
    ISA_Bus->write_io_8(AdresReg,0x38);
    ISA_Bus->write_io_8(DataReg,0x03);


	//VGA reg $21... (enable reprogram!)
    ISA_Bus->write_io_8(AdresReg,0x21);
 	ISA_Bus->write_io_8(DataReg,0xfa);

	//SEQ:
    ISA_Bus->write_io_8(0x03c4,0x01);
    ISA_Bus->write_io_8(0x03c5,0x01);


	//Set overscan color to 'black':
	//Disable this part if you're trying to center the output on TV,
	//you'll get blue overscan range color then. Use this as a guide-'line' ;-)
	//(select index adress register:)
	ISA_Bus->read_io_8(0x3da);
	//(write index for 'overscan color' register:)
    ISA_Bus->write_io_8(0x03c0,0x11);
    //(write data for 'overscan color' register:)
    ISA_Bus->write_io_8(0x03c0,0x00);


	//unlock VGA registers at index 0-7
    ISA_Bus->write_io_8(AdresReg,0x11);
	temp = (ISA_Bus->read_io_8(DataReg) & 0x7f);
	ISA_Bus->write_io_8(DataReg,temp);

	//unlock VGA registers 'vert. retrace start' & 'vert. retrace end':
    ISA_Bus->write_io_8(AdresReg,0x03);
	ISA_Bus->write_io_8(DataReg,
		(ISA_Bus->read_io_8(DataReg) | 0x80));

   	//'update' PIXEL/TV:
    ISA_Bus->write_io_8(AdresReg,0x28);
	temp = (ISA_Bus->read_io_8(DataReg) & 0xc7);
 	ISA_Bus->write_io_8(DataReg,temp | 0x80);

  	//LCD:
    ISA_Bus->write_io_8(AdresReg,0x33);
	temp = (ISA_Bus->read_io_8(DataReg) & 0xfe);
    ISA_Bus->write_io_8(AdresReg,0x33);
 	ISA_Bus->write_io_8(DataReg,temp);

	//TREG:
	ISA_Bus->write_io_8(AdresReg,0x3d);
	ISA_Bus->write_io_8(DataReg,0x80);
	LEAVE_CRITICAL_SECTION
}


//IIC implementation (bit-banger algorithm):

char FlagIICError (char ErrNo)
//error code list:
//1 - SCL locked low by device (bus is still busy)
//2 - SDA locked low by device (bus is still busy)
//3 - No Acknowledge from device (no handshake)
//4 - SDA not released for master to generate STOP bit
{
	static char IICError = 0;

	if (!IICError) IICError = ErrNo;
	if (ErrNo == -1) IICError = 0;
	return IICError;
}

void bstart (int BusNR)
{
	dprintf ("riva IICbus: START condition generated on bus %d.\n",BusNR);
	//spinlock..
	OutSDA(1,BusNR); //make sure SDA is high
	snooze(2);
	OutSCL(1,BusNR);
	snooze(2); //'wait' 2uS
	if (!InSCL(BusNR)) FlagIICError (1);
	snooze(5);
	//clear SDA while SCL set (bus-start condition):
	OutSDA(0,BusNR);
	snooze(5);
	OutSCL(0,BusNR);
	snooze(5);
	//clear spinlock..
}

void bstop (int BusNR)
{
	dprintf ("riva IICbus: STOP condition generated on bus %d.\n",BusNR);
	//spinlock..
	OutSDA(0,BusNR); //make sure SDA is low
	snooze(2);
	OutSCL(1,BusNR);
	snooze(2);
	if (!InSCL(BusNR)) FlagIICError (1);
	snooze(5);
	//set SDA while SCL set (bus-stop condition):
	OutSDA(1,BusNR);
	snooze(2);
	if (!InSDA(BusNR)) FlagIICError (4);
	snooze(2);
	//clear spinlock..
}

void TXBit (char Bit, int BusNR)
{
	//spinlock..
	//send out databit:
	if (Bit)
	{
		OutSDA(1, BusNR);
		dprintf ("1");
		spin(2);
		if (!InSDA(BusNR)) FlagIICError (2);
	}
	else
	{
		OutSDA(0, BusNR);
		dprintf ("0");
	}
	//generate clock pulse:
	spin(5);
	OutSCL(1,BusNR);
	spin(2);
	if (!InSCL(BusNR)) FlagIICError (1);
	spin(5);
	OutSCL(0,BusNR);
	spin(5);
	//clear spinlock..
}

char RXBit (int BusNR)
{
	char Bit = 0;
	
	//spinlock..
	OutSDA(1,BusNR); //set SDA so input is possible
	//generate clock pulse:
	spin(5);
	OutSCL(1,BusNR);
	spin(2);
	if (!InSCL(BusNR)) FlagIICError (1);
	spin(2);
	//read in databit:
	if (InSDA(BusNR))
	{
		Bit = 1;
		dprintf ("1");
	}
	else
	{
		dprintf ("0");
	}
	OutSCL(0,BusNR);
	spin(5);
	//clear spinlock..
	return Bit;
}

void TXByte (char Byte, int BusNR)
//transmit one byte starting with MSbit.
//Afterwards check for low ACK signal from slave device.
{
	unsigned char BitCount, Bit;
	
	dprintf ("riva IICbus %d: TX %%",BusNR);
	for (BitCount = 8; BitCount > 0; BitCount--)
	{
		Bit = (Byte & 0x80);
		TXBit (Bit, BusNR);
		Byte <<= 1;
	}
	dprintf (", RX ACK=");
	Bit = RXBit (BusNR);
	if (Bit) FlagIICError (3);
	dprintf (", bus status %d.\n", FlagIICError(0));

}

char RXByte (char Ack, int BusNR)
//receive one byte starting with MSbit.
//Afterwards send (low or high) ACK signal to slave device.
{
	unsigned char BitCount, Bit;
	char Byte = 0;
	
	dprintf ("riva IICbus %d: RX %%",BusNR);
	for (BitCount = 8; BitCount > 0; BitCount--)
	{
		Byte <<= 1;
		Bit = RXBit (BusNR);
		Byte += Bit;
	}
	dprintf (", TX ACK=");
	TXBit (Ack, BusNR);
	dprintf (", bus status %d.\n", FlagIICError(0));
	return Byte;
}

//low-level hardware manipulation functions:

void OutSCL(char Bit, int BusNR)
{
	unsigned char Data;
	
	//spinlock...
	ENTER_CRITICAL_SECTION
	if (BusNR)
	{
		ISA_Bus->write_io_8(AdresReg,WRIICBus1);
	}
	else
	{
		ISA_Bus->write_io_8(AdresReg,WRIICBus0);
	}

	Data = ISA_Bus->read_io_8(DataReg) & 0xf0;
	if (Bit)
		Data |= 0x20;
	else
		Data &= ~0x20;

	if (BusNR)
	{
		ISA_Bus->write_io_8(AdresReg,WRIICBus1);
	}
	else
	{
		ISA_Bus->write_io_8(AdresReg,WRIICBus0);
	}

	ISA_Bus->write_io_8(DataReg,Data | 0x01);
	//end spinlock...
	LEAVE_CRITICAL_SECTION
}

void OutSDA(char Bit, int BusNR)
{
	unsigned char Data;
	
	//spinlock...
	ENTER_CRITICAL_SECTION
	if (BusNR)
	{
		ISA_Bus->write_io_8(AdresReg,WRIICBus1);
	}
	else
	{
		ISA_Bus->write_io_8(AdresReg,WRIICBus0);
	}

	Data = ISA_Bus->read_io_8(DataReg) & 0xf0;
	if (Bit)
		Data |= 0x10;
	else
		Data &= ~0x10;

	if (BusNR)
	{
		ISA_Bus->write_io_8(AdresReg,WRIICBus1);
	}
	else
	{
		ISA_Bus->write_io_8(AdresReg,WRIICBus0);
	}

	ISA_Bus->write_io_8(DataReg,Data | 0x01);
	//end spinlock...
	LEAVE_CRITICAL_SECTION
}

char InSCL(int BusNR)
{
	unsigned char Data = 0;
	
	//spinlock
	ENTER_CRITICAL_SECTION
	if (BusNR)
	{
		ISA_Bus->write_io_8(AdresReg,RDIICBus1);
	}
	else
	{
		ISA_Bus->write_io_8(AdresReg,RDIICBus0);
	}

	if (ISA_Bus->read_io_8(DataReg) & 0x04)
		Data = 1;
	//end spinlock...
	LEAVE_CRITICAL_SECTION
	return Data;
}

char InSDA(int BusNR)
{
	unsigned char Data = 0;
	
	//spinlock
	ENTER_CRITICAL_SECTION
	if (BusNR)
	{
		ISA_Bus->write_io_8(AdresReg,RDIICBus1);
	}
	else
	{
		ISA_Bus->write_io_8(AdresReg,RDIICBus0);
	}

	if (ISA_Bus->read_io_8(DataReg) & 0x08)
		Data = 1;
	//end spinlock...
	LEAVE_CRITICAL_SECTION
	return Data;
}